@@ -26,6 +26,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <sched.h>
 #include <limits.h>
 
 struct ct_general_state st;
@@ -43,7 +44,7 @@ static const char usage_client_commands[] =
 	"  -i, display content of the internal cache\n"
 	"  -e, display the content of the external cache\n"
 	"  -k, kill conntrack daemon\n"
-	"  -s  [|network|cache|runtime|link|queue], dump statistics\n"
+	"  -s  [|network|cache|runtime|link|rsqueue|queue], dump statistics\n"
 	"  -R, resync with kernel conntrack table\n"
 	"  -n, request resync with other node (only FT-FW and NOTRACK modes)\n"
 	"  -x, dump cache in XML format (requires -i or -e)\n"
@@ -209,6 +210,14 @@ int main(int argc, char *argv[])
 						 strlen(argv[i+1])) == 0) {
 					action = STATS_LINK;
 					i++;
+				} else if (strncmp(argv[i+1], "rsqueue",
+						strlen(argv[i+1])) == 0) {
+					action = STATS_RSQUEUE;
+					i++;
+				} else if (strncmp(argv[i+1], "process",
+						 strlen(argv[i+1])) == 0) {
+					action = STATS_PROCESS;
+					i++;
 				} else if (strncmp(argv[i+1], "queue",
 						strlen(argv[i+1])) == 0) {
 					action = STATS_QUEUE;
@@ -295,6 +304,23 @@ int main(int argc, char *argv[])
 	}
 	close(ret);
 
+	/*
+	 * Setting process priority and scheduler
+	 */
+	nice(CONFIG(nice));
+
+	if (CONFIG(sched).type != SCHED_OTHER) {
+		struct sched_param schedparam = {
+			.sched_priority = CONFIG(sched).prio,
+		};
+
+		ret = sched_setscheduler(0, CONFIG(sched).type, &schedparam);
+		if (ret == -1) {
+			perror("sched");
+			exit(EXIT_FAILURE);
+		}
+	}
+
 	/*
 	 * initialization process
 	 */
@@ -309,7 +335,6 @@ int main(int argc, char *argv[])
 
 	chdir("/");
 	close(STDIN_FILENO);
-	nice(CONFIG(nice));
 
 	/* Daemonize conntrackd */
 	if (type == DAEMON) {
